@@ -37,6 +37,7 @@ from The Open Group.
 #include "xauth.h"
 #include <ctype.h>
 #include <errno.h>
+#include <stdint.h>
 #include <sys/stat.h>
 #ifndef WIN32
 #include <sys/socket.h>
@@ -251,6 +252,18 @@ skip_nonspace(register char *s)
     return s;
 }
 
+#ifndef HAVE_REALLOCARRAY
+static inline void *
+reallocarray(void *optr, size_t nmemb, size_t size)
+{
+    if ((nmemb > 0) && (SIZE_MAX / nmemb < size)) {
+        errno = ENOMEM;
+        return NULL;
+    }
+    return realloc(optr, size * nmemb);
+}
+#endif
+
 static const char **
 split_into_words(char *src, int *argcp)  /* argvify string */
 {
@@ -278,9 +291,15 @@ split_into_words(char *src, int *argcp)  /* argvify string */
 	savec = *src;
 	*src = '\0';
 	if (cur == total) {
+	    const char **new_argv;
 	    total += WORDSTOALLOC;
-	    argv = realloc (argv, total * sizeof (char *));
-	    if (!argv) return NULL;
+	    new_argv = reallocarray (argv, total, sizeof (char *));
+	    if (new_argv != NULL) {
+		argv = new_argv;
+	    } else {
+		free(argv);
+		return NULL;
+	    }
 	}
 	argv[cur++] = jword;
 	if (savec) src++;		/* if not last on line advance */
@@ -633,18 +652,19 @@ static Bool xauth_modified = False;	/* if added, removed, or merged */
 static Bool xauth_allowed = True;	/* if allowed to write auth file */
 static Bool xauth_locked = False;     /* if has been locked */
 static const char *xauth_filename = NULL;
-static volatile Bool dieing = False;
+static volatile Bool dying = False;
 
 
-/* poor man's puts(), for under signal handlers */
-#define WRITES(fd, S) (void)write((fd), (S), strlen((S)))
+/* poor man's puts(), for under signal handlers, 
+   extended to ignore warn_unused_result */
+#define WRITES(fd, S) {if(write((fd), (S), strlen((S))));}
 
 /* ARGSUSED */
 _X_NORETURN
 static void
 die(int sig)
 {
-    dieing = True;
+    dying = True;
     _exit (auth_finalize ());
     /* NOTREACHED */
 }
@@ -696,6 +716,10 @@ auth_initialize(const char *authfilename)
     FILE *authfp;
     Bool exists;
 
+    if (strlen(authfilename) > 1022) {
+	fprintf (stderr, "%s: authority file name \"%s\" too long\n",
+		 ProgramName, authfilename);
+    }
     xauth_filename = authfilename;    /* used in cleanup, prevent race with
                                          signals */
     register_signals ();
@@ -843,17 +867,20 @@ write_auth_file(char *tmp_nam)
 	}
     }
 
-    (void) fclose (fp);
+    if (fclose(fp)) {
+	return -1;
+    }
+
     return 0;
 }
 
 int
 auth_finalize(void)
 {
-    char temp_name[1024];	/* large filename size */
+    char temp_name[1025];	/* large filename size */
 
     if (xauth_modified) {
-	if (dieing) {
+	if (dying) {
 	    if (verbose) {
 		/*
 		 * called from a signal handler -- printf is *not* reentrant; also
@@ -886,19 +913,11 @@ auth_finalize(void)
 			 "%s:  unable to write authority file %s\n",
 			 ProgramName, temp_name);
 	    } else {
-		(void) unlink (xauth_filename);
-#if defined(WIN32) || defined(__UNIXOS2__)
-		if (rename(temp_name, xauth_filename) == -1)
-#else
-		/* Attempt to rename() if link() fails, since this may be on a FS that does not support hard links */
-		if (link (temp_name, xauth_filename) == -1 && rename(temp_name, xauth_filename) == -1)
-#endif
-		{
+		if (rename(temp_name, xauth_filename) == -1) {
 		    fprintf (stderr,
-		     "%s:  unable to link authority file %s, use %s\n",
+		     "%s:  unable to rename authority file %s, use %s\n",
 			     ProgramName, xauth_filename, temp_name);
-		} else {
-		    (void) unlink (temp_name);
+                    unlink(temp_name);
 		}
 	    }
 	}
@@ -1004,7 +1023,7 @@ dump_entry(const char *inputfilename, int lineno, Xauth *auth, char *data)
 		fprintf (fp, "%s", dpyname);
 		break;
 	    }
-	    /* else fall through to default */
+	    /* else fall through */
 	  default:
 	    fprintf (fp, "#%04x#", auth->family);
 	    fprintfhex (fp, auth->address_length, auth->address);
@@ -1053,16 +1072,22 @@ extract_entry(const char *inputfilename, int lineno, Xauth *auth, char *data)
 
 
 static int
-eq_auth(Xauth *a, Xauth *b)
+eq_auth_dpy_and_name(Xauth *a, Xauth *b)
 {
     return((a->family == b->family &&
 	    a->address_length == b->address_length &&
 	    a->number_length == b->number_length &&
 	    a->name_length == b->name_length &&
-	    a->data_length == b->data_length &&
 	    memcmp(a->address, b->address, a->address_length) == 0 &&
 	    memcmp(a->number, b->number, a->number_length) == 0 &&
-	    memcmp(a->name, b->name, a->name_length) == 0 &&
+	    memcmp(a->name, b->name, a->name_length) == 0) ? 1 : 0);
+}
+
+static int
+eq_auth(Xauth *a, Xauth *b)
+{
+    return((eq_auth_dpy_and_name(a, b) &&
+	    a->data_length == b->data_length &&
 	    memcmp(a->data, b->data, a->data_length) == 0) ? 1 : 0);
 }
 
@@ -1096,17 +1121,6 @@ match_auth_dpy(register Xauth *a, register Xauth *b)
     return 1;
 }
 
-/* return non-zero iff display and authorization type are the same */
-
-static int
-match_auth(register Xauth *a, register Xauth *b)
-{
-    return ((match_auth_dpy(a, b)
-	     && a->name_length == b->name_length
-	     && memcmp(a->name, b->name, a->name_length) == 0) ? 1 : 0);
-}
-
-
 static int
 merge_entries(AuthList **firstp, AuthList *second, int *nnewp, int *nreplp)
 {
@@ -1140,7 +1154,7 @@ merge_entries(AuthList **firstp, AuthList *second, int *nnewp, int *nreplp)
 
 	a = first;
 	for (;;) {
-	    if (match_auth (a->auth, b->auth)) {  /* found a duplicate */
+	    if (eq_auth_dpy_and_name (a->auth, b->auth)) {  /* found a duplicate */
 		AuthList tmp;		/* swap it in for old one */
 		tmp = *a;
 		*a = *b;
@@ -1171,6 +1185,45 @@ merge_entries(AuthList **firstp, AuthList *second, int *nnewp, int *nreplp)
 
 }
 
+static void
+sort_entries(AuthList **firstp)
+{
+    /* Insert sort, in each pass it removes auth records of certain */
+    /* cathegory from the given list and inserts them into the sorted list. */
+
+    AuthList *sorted = NULL, *sorted_tail = NULL;
+    AuthList *prev, *iter, *next;
+
+    #define SORT_OUT(EXPRESSION) { \
+	prev = NULL; \
+	for (iter = *firstp; iter; iter = next) { \
+	    next = iter->next; \
+	    if (EXPRESSION) { \
+		if (prev) \
+		    prev->next = next; \
+		else \
+		    *firstp = next; \
+		if (sorted_tail == NULL) { \
+		    sorted = sorted_tail = iter; \
+		} else { \
+		    sorted_tail->next = iter; \
+		    sorted_tail = iter; \
+		} \
+		iter->next = NULL; \
+	    } else { \
+		prev = iter; \
+	    } \
+	} \
+    }
+
+    SORT_OUT(iter->auth->family != FamilyWild && iter->auth->number_length != 0);
+    SORT_OUT(iter->auth->family != FamilyWild && iter->auth->number_length == 0);
+    SORT_OUT(iter->auth->family == FamilyWild && iter->auth->number_length != 0);
+    SORT_OUT(iter->auth->family == FamilyWild && iter->auth->number_length == 0);
+
+    *firstp = sorted;
+}
+
 static Xauth *
 copyAuth(Xauth *auth)
 {
@@ -1509,6 +1562,7 @@ do_merge(const char *inputfilename, int lineno, int argc, const char **argv)
 	  printf ("%d entries read in:  %d new, %d replacement%s\n",
 	  	  nentries, nnew, nrepl, nrepl != 1 ? "s" : "");
 	if (nentries > 0) xauth_modified = True;
+	sort_entries(&xauth_head);
     }
 
     return 0;
@@ -1583,13 +1637,22 @@ do_add(const char *inputfilename, int lineno, int argc, const char **argv)
     hexkey = argv[3];
 
     len = strlen(hexkey);
-    if (hexkey[0] == '"' && hexkey[len-1] == '"') {
+    if (len > 1 && hexkey[0] == '"' && hexkey[len-1] == '"') {
 	key = malloc(len-1);
+	if (!key) {
+	    fprintf(stderr, "unable to allocate memory\n");
+	    return 1;
+	}
 	strncpy(key, hexkey+1, len-2);
+	key[len-1] = '\0';
 	len -= 2;
     } else if (!strcmp(protoname, SECURERPC) ||
 	       !strcmp(protoname, K5AUTH)) {
 	key = malloc(len+1);
+	if (!key) {
+	    fprintf(stderr, "unable to allocate memory\n");
+	    return 1;
+	}
 	strcpy(key, hexkey);
     } else {
 	len = cvthexkey (hexkey, &key);
@@ -1657,6 +1720,7 @@ do_add(const char *inputfilename, int lineno, int argc, const char **argv)
 	fprintf (stderr, "unable to merge in added record\n");
 	return 1;
     }
+    sort_entries(&xauth_head);
 
     xauth_modified = True;
     return 0;
@@ -1827,10 +1891,10 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
     const char *displayname;
     int major_version, minor_version;
     XSecurityAuthorization id_return;
-    Xauth *auth_in, *auth_return;
+    Xauth *auth_in = NULL, *auth_return = NULL;
     XSecurityAuthorizationAttributes attributes;
     unsigned long attrmask = 0;
-    Display *dpy;
+    Display *dpy = NULL;
     int status;
     const char *args[4];
     const char *protoname = ".";
@@ -1838,7 +1902,7 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
     int authdatalen = 0;
     const char *hexdata;
     char *authdata = NULL;
-    char *hex;
+    char *hex = NULL;
 
     if (argc < 2 || !argv[1]) {
 	prefix (inputfilename, lineno);
@@ -1857,7 +1921,8 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
 	    if (++i == argc) {
 		prefix (inputfilename, lineno);
 		badcommandline (argv[i-1]);
-		return 1;
+		status = 1;
+		goto exit_generate;
 	    }
 	    attributes.timeout = atoi(argv[i]);
 	    attrmask |= XSecurityTimeout;
@@ -1874,7 +1939,8 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
 	    if (++i == argc) {
 		prefix (inputfilename, lineno);
 		badcommandline (argv[i-1]);
-		return 1;
+		status = 1;
+		goto exit_generate;
 	    }
 	    attributes.group = atoi(argv[i]);
 	    attrmask |= XSecurityGroup;
@@ -1883,13 +1949,20 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
 	    if (++i == argc) {
 		prefix (inputfilename, lineno);
 		badcommandline (argv[i-1]);
-		return 1;
+		status = 1;
+		goto exit_generate;
 	    }
 	    hexdata = argv[i];
 	    authdatalen = strlen(hexdata);
 	    if (hexdata[0] == '"' && hexdata[authdatalen-1] == '"') {
 		authdata = malloc(authdatalen-1);
+		if (!authdata) {
+		    fprintf(stderr, "unable to allocate memory\n");
+		    status = 1;
+		    goto exit_generate;
+		}
 		strncpy(authdata, hexdata+1, authdatalen-2);
+		authdata[authdatalen-1] = '\0';
 		authdatalen -= 2;
 	    } else {
 		authdatalen = cvthexkey (hexdata, &authdata);
@@ -1897,13 +1970,15 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
 		    prefix (inputfilename, lineno);
 		    fprintf (stderr,
 			     "data contains odd number of or non-hex characters\n");
-		    return 1;
+		    status = 1;
+		    goto exit_generate;
 		}
 	    }
 	} else {
 	    prefix (inputfilename, lineno);
 	    badcommandline (argv[i]);
-	    return 1;
+	    status = 1;
+	    goto exit_generate;
 	}
     }
 
@@ -1913,7 +1988,8 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
     if (!dpy) {
 	prefix (inputfilename, lineno);
 	fprintf (stderr, "unable to open display \"%s\".\n", displayname);
-	return 1;
+	status = 1;
+	goto exit_generate;
     }
 
     status = XSecurityQueryExtension(dpy, &major_version, &minor_version);
@@ -1922,7 +1998,8 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
 	prefix (inputfilename, lineno);
 	fprintf (stderr, "couldn't query Security extension on display \"%s\"\n",
 		 displayname);
-        return 1;
+	status = 1;
+	goto exit_generate;
     }
 
     /* fill in input Xauth struct */
@@ -1947,7 +2024,8 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
     {
 	prefix (inputfilename, lineno);
 	fprintf (stderr, "couldn't generate authorization\n");
-	return 1;
+	status = 1;
+	goto exit_generate;
     }
 
     if (verbose)
@@ -1962,10 +2040,12 @@ do_generate(const char *inputfilename, int lineno, int argc, const char **argv)
 
     status = do_add(inputfilename, lineno, 4, args);
 
-    if (authdata) free(authdata);
+  exit_generate:
+    free(authdata);
     XSecurityFreeXauth(auth_in);
     XSecurityFreeXauth(auth_return);
     free(hex);
-    XCloseDisplay(dpy);
+    if (dpy != NULL)
+	XCloseDisplay(dpy);
     return status;
 }
